import unittest
from unittest import mock

from flumine import config
from flumine.controls.tradingcontrols import (
    OrderValidation,
    StrategyExposure,
    OrderTypes,
    ExchangeType,
    OrderPackageType,
    MarketValidation,
    ExecutionValidation,
)
from flumine.markets.blotter import Blotter
from flumine.order.order import OrderStatus
from flumine.streams.orderstream import OrderStream


class TestOrderValidation(unittest.TestCase):
    def setUp(self):
        self.mock_flumine = mock.Mock()
        self.mock_client = mock.Mock()
        self.mock_client.min_bet_size = 2
        self.mock_client.min_bsp_liability = 10
        self.trading_control = OrderValidation(self.mock_flumine)

    def test_init(self):
        assert self.trading_control.NAME == "ORDER_VALIDATION"

    @mock.patch(
        "flumine.controls.tradingcontrols.OrderValidation._validate_betfair_order"
    )
    def test_validate_betfair(self, mock_validate_betfair_order):
        mock_order = mock.Mock(EXCHANGE=ExchangeType.BETFAIR)
        mock_order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        mock_order.order_type.size = 12
        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        mock_validate_betfair_order.assert_called_with(mock_order)

    @mock.patch(
        "flumine.controls.tradingcontrols.OrderValidation._validate_betdaq_order"
    )
    def test_validate_betdaq(self, mock_validate_betfair_order):
        mock_order = mock.Mock(EXCHANGE=ExchangeType.BETDAQ)
        mock_order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        mock_order.order_type.size = 12
        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        mock_validate_betfair_order.assert_called_with(mock_order)

    @mock.patch(
        "flumine.controls.tradingcontrols.OrderValidation._validate_betfair_min_size"
    )
    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._validate_size")
    @mock.patch(
        "flumine.controls.tradingcontrols.OrderValidation._validate_betfair_price"
    )
    def test__validate_betfair_order_limit(
        self,
        mock__validate_betfair_price,
        mock__validate_size,
        mock__validate_betfair_min_size,
    ):
        order = mock.Mock()
        order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        self.trading_control._validate_betfair_order(order)
        mock__validate_betfair_price.assert_called_with(order)
        mock__validate_size.assert_called_with(order)
        mock__validate_betfair_min_size.assert_called_with(order, OrderTypes.LIMIT)

    @mock.patch(
        "flumine.controls.tradingcontrols.OrderValidation._validate_betfair_price"
    )
    @mock.patch(
        "flumine.controls.tradingcontrols.OrderValidation._validate_betfair_min_size"
    )
    @mock.patch(
        "flumine.controls.tradingcontrols.OrderValidation._validate_betfair_liability"
    )
    def test__validate_betfair_order_limit_on_close(
        self,
        mock__validate_betfair_liability,
        mock__validate_betfair_min_size,
        mock__validate_betfair_price,
    ):
        order = mock.Mock()
        order.order_type.ORDER_TYPE = OrderTypes.LIMIT_ON_CLOSE
        self.trading_control._validate_betfair_order(order)
        mock__validate_betfair_liability.assert_called_with(order)
        mock__validate_betfair_min_size.assert_called_with(
            order, OrderTypes.LIMIT_ON_CLOSE
        )
        mock__validate_betfair_price.assert_called_with(order)

    @mock.patch(
        "flumine.controls.tradingcontrols.OrderValidation._validate_betfair_min_size"
    )
    @mock.patch(
        "flumine.controls.tradingcontrols.OrderValidation._validate_betfair_liability"
    )
    def test__validate_betfair_order_market_on_close(
        self, mock__validate_betfair_liability, mock__validate_betfair_min_size
    ):
        order = mock.Mock()
        order.order_type.ORDER_TYPE = OrderTypes.MARKET_ON_CLOSE
        self.trading_control._validate_betfair_order(order)
        mock__validate_betfair_liability.assert_called_with(order)
        mock__validate_betfair_min_size.assert_called_with(
            order, OrderTypes.MARKET_ON_CLOSE
        )

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_order_unknown(self, mock_on_error):
        order = mock.Mock()
        order.order_type.ORDER_TYPE = "RUSH"
        self.trading_control._validate_betfair_order(order)
        mock_on_error.assert_called_with(order, "Unknown orderType")

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_size(self, mock_on_error):
        order = mock.Mock()
        order.order_type.size = 2
        self.trading_control._validate_size(order)
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_size_on_error(self, mock_on_error):
        order = mock.Mock()
        order.order_type.size = None
        order.order_type.bet_target_size = None
        self.trading_control._validate_size(order)
        mock_on_error.assert_called_with(order, "Order size is None")

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_size_on_error_two(self, mock_on_error):
        order = mock.Mock()
        order.order_type.size = -1
        self.trading_control._validate_size(order)
        mock_on_error.assert_called_with(order, "Order size is less than 0")

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_target_size_on_error_two(self, mock_on_error):
        order = mock.Mock(EXCHANGE=ExchangeType.BETFAIR)
        order.order_type.size = None
        order.order_type.bet_target_size = -1
        self.trading_control._validate_size(order)
        mock_on_error.assert_called_with(order, "Order size is less than 0")

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_size_on_error_three(self, mock_on_error):
        order = mock.Mock()
        order.order_type.size = 1.999
        self.trading_control._validate_size(order)
        mock_on_error.assert_called_with(order, "Order size has more than 2dp")

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_target_size_on_error_three(self, mock_on_error):
        order = mock.Mock(EXCHANGE=ExchangeType.BETFAIR)
        order.order_type.size = None
        order.order_type.bet_target_size = 1.999
        self.trading_control._validate_size(order)
        mock_on_error.assert_called_with(order, "Order size has more than 2dp")

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_price(self, mock_on_error):
        order = mock.Mock()
        order.order_type.price = 2
        order.order_type.price_ladder_definition = "CLASSIC"
        self.trading_control._validate_betfair_price(order)
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_price_on_error(self, mock_on_error):
        order = mock.Mock()
        order.order_type.price = None
        self.trading_control._validate_betfair_price(order)
        mock_on_error.assert_called_with(order, "Order price is None")

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_price_on_error_two(self, mock_on_error):
        order = mock.Mock()
        order.order_type.price = -1
        order.order_type.price_ladder_definition = "CLASSIC"
        self.trading_control._validate_betfair_price(order)
        mock_on_error.assert_called_with(
            order, "Order price is not valid for CLASSIC ladder"
        )

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_price_finest(self, mock_on_error):
        order = mock.Mock()
        order.order_type.price = 999.01
        order.order_type.price_ladder_definition = "FINEST"
        self.trading_control._validate_betfair_price(order)
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_price_finest_error(self, mock_on_error):
        order = mock.Mock()
        order.order_type.price = 1000.01
        order.order_type.price_ladder_definition = "FINEST"
        self.trading_control._validate_betfair_price(order)
        mock_on_error.assert_called_with(
            order, "Order price is not valid for FINEST ladder"
        )

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_price_line_range(self, mock_on_error):
        order = mock.Mock()
        order.order_type.price = 999.5
        order.order_type.price_ladder_definition = "LINE_RANGE"
        order.order_type.line_range_info = mock.Mock(
            min_unit_value=-0.5, max_unit_value=999.5, interval=1.0
        )
        self.trading_control._validate_betfair_price(order)
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_price_line_range_error(self, mock_on_error):
        order = mock.Mock()
        order.order_type.price = 1000.0
        order.order_type.price_ladder_definition = "LINE_RANGE"
        order.order_type.line_range_info = mock.Mock(
            min_unit_value=-0.5, max_unit_value=999.5, interval=1.0
        )
        self.trading_control._validate_betfair_price(order)
        mock_on_error.assert_called_with(
            order, "Order price is not valid for LINE_RANGE ladder"
        )

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_liability(self, mock_on_error):
        order = mock.Mock()
        order.order_type.liability = 2
        self.trading_control._validate_betfair_liability(order)
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_liability_on_error(self, mock_on_error):
        order = mock.Mock()
        order.order_type.liability = None
        self.trading_control._validate_betfair_liability(order)
        mock_on_error.assert_called_with(order, "Order liability is None")

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_liability_on_error_two(self, mock_on_error):
        order = mock.Mock()
        order.order_type.liability = -1
        self.trading_control._validate_betfair_liability(order)
        mock_on_error.assert_called_with(order, "Order liability is less than 0")

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_liability_on_error_three(self, mock_on_error):
        order = mock.Mock()
        order.order_type.liability = 2.999
        self.trading_control._validate_betfair_liability(order)
        mock_on_error.assert_called_with(order, "Order liability has more than 2dp")

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_min_size_no_validation(self, mock_on_error):
        mock_client = mock.Mock()
        mock_client.min_bet_validation = False
        order = mock.Mock(client=mock_client, side="BACK")
        order.order_type.size = 0.01
        order.order_type.price = 2
        self.trading_control._validate_betfair_min_size(order, OrderTypes.LIMIT)
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_min_size_limit(self, mock_on_error):
        mock_client = mock.Mock(min_bet_size=2, min_bet_payout=10)
        order = mock.Mock(client=mock_client)
        order.side = "BACK"
        order.order_type.size = 2
        order.order_type.price = 2
        self.trading_control._validate_betfair_min_size(order, OrderTypes.LIMIT)
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_min_size_limit_large(self, mock_on_error):
        mock_client = mock.Mock(min_bet_size=0.10, min_bet_payout=100)
        order = mock.Mock(client=mock_client)
        order.side = "BACK"
        order.order_type.size = 2
        order.order_type.price = 2
        self.trading_control._validate_betfair_min_size(order, OrderTypes.LIMIT)
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_min_size_limit_error(self, mock_on_error):
        mock_client = mock.Mock(min_bet_size=2, min_bet_payout=10)
        order = mock.Mock(client=mock_client)
        order.side = "BACK"
        order.order_type.size = 1.99
        order.order_type.price = 2
        self.trading_control._validate_betfair_min_size(order, OrderTypes.LIMIT)
        mock_on_error.assert_called_with(
            order,
            "Order size is less than min bet size (2) or payout (10) for currency",
        )

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_min_target_size_limit_error(self, mock_on_error):
        mock_client = mock.Mock(min_bet_size=2, min_bet_payout=10)
        order = mock.Mock(client=mock_client)
        order.side = "BACK"
        order.order_type.size = None
        order.order_type.bet_target_size = 1.99
        order.order_type.price = 2
        self.trading_control._validate_betfair_min_size(order, OrderTypes.LIMIT)
        mock_on_error.assert_called_with(
            order,
            "Order size is less than min bet size (2) or payout (10) for currency",
        )

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_min_size(self, mock_on_error):
        mock_client = mock.Mock(min_bet_size=2)
        order = mock.Mock(client=mock_client)
        order.side = "BACK"
        order.order_type.liability = 2
        self.trading_control._validate_betfair_min_size(
            order, OrderTypes.LIMIT_ON_CLOSE
        )
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_min_size_two(self, mock_on_error):
        mock_client = mock.Mock(min_bsp_liability=10)
        order = mock.Mock(client=mock_client)
        order.side = "LAY"
        order.order_type.liability = 10
        self.trading_control._validate_betfair_min_size(
            order, OrderTypes.LIMIT_ON_CLOSE
        )
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_min_size_on_error(self, mock_on_error):
        mock_client = mock.Mock(min_bet_size=2)
        order = mock.Mock(client=mock_client)
        order.side = "BACK"
        order.order_type.liability = 1.99
        self.trading_control._validate_betfair_min_size(
            order, OrderTypes.LIMIT_ON_CLOSE
        )
        mock_on_error.assert_called_with(
            order, "Liability is less than min bet size (2) for currency"
        )

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betfair_min_size_on_error_two(self, mock_on_error):
        mock_client = mock.Mock(min_bsp_liability=10)
        order = mock.Mock(client=mock_client)
        order.side = "LAY"
        order.order_type.liability = 9.99
        self.trading_control._validate_betfair_min_size(
            order, OrderTypes.LIMIT_ON_CLOSE
        )
        mock_on_error.assert_called_with(
            order, "Liability is less than min BSP payout (10) for currency"
        )

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betdaq_price(self, mock_on_error):
        order = mock.Mock()
        order.order_type.price = 2
        self.trading_control._validate_betdaq_price(order)
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betdaq_price_error(self, mock_on_error):
        order = mock.Mock()
        order.order_type.price = 501
        self.trading_control._validate_betdaq_price(order)
        mock_on_error.assert_called_with(
            order, "Order price is not valid for BETDAQ ladder"
        )

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betdaq_min_size(self, mock_on_error):
        mock_client = mock.Mock(min_bet_size=2)
        order = mock.Mock(client=mock_client)
        order.side = "BACK"
        order.order_type.size = 2
        self.trading_control._validate_betdaq_min_size(order, OrderTypes.LIMIT_ON_CLOSE)
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.OrderValidation._on_error")
    def test__validate_betdaq_min_size_error(self, mock_on_error):
        mock_client = mock.Mock(min_bet_size=2)
        order = mock.Mock(client=mock_client)
        order.side = "BACK"
        order.order_type.size = 0.09
        self.trading_control._validate_betdaq_min_size(order, OrderTypes.LIMIT_ON_CLOSE)
        mock_on_error.assert_called()


class TestMarketValidation(unittest.TestCase):
    def setUp(self):
        self.mock_market = mock.Mock(market_book=mock.Mock(status="OPEN"))
        self.mock_flumine = mock.Mock()
        self.mock_flumine.markets.markets = {"market_id": self.mock_market}
        self.trading_control = MarketValidation(self.mock_flumine)
        self.mock_order = mock.Mock()
        self.mock_order.market_id = "market_id"

    @mock.patch("flumine.controls.tradingcontrols.MarketValidation._on_error")
    def test__validate_betfair_market_status(self, mock_on_error):
        self.trading_control._validate_betfair_market_status(self.mock_order)
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.MarketValidation._on_error")
    def test__validate_betfair_market_status_closed(self, mock_on_error):
        self.mock_market.market_book.status = "CLOSED"
        self.trading_control._validate_betfair_market_status(self.mock_order)
        mock_on_error.assert_called()

    @mock.patch("flumine.controls.tradingcontrols.MarketValidation._on_error")
    def test__validate_betfair_market_status_suspended(self, mock_on_error):
        self.mock_market.market_book.status = "SUSPENDED"
        self.trading_control._validate_betfair_market_status(self.mock_order)
        mock_on_error.assert_called()

    @mock.patch("flumine.controls.tradingcontrols.MarketValidation._on_error")
    def test__validate_betfair_market_not_found(self, mock_on_error):
        self.mock_flumine.markets.markets = {}
        self.trading_control._validate_betfair_market_status(self.mock_order)
        mock_on_error.assert_called()

    @mock.patch("flumine.controls.tradingcontrols.MarketValidation._on_error")
    def test__validate_betfair_market_book_not_available(self, mock_on_error):
        self.mock_market.market_book = None
        self.trading_control._validate_betfair_market_status(self.mock_order)
        mock_on_error.assert_called()


class TestExecutionValidation(unittest.TestCase):
    def setUp(self):
        self.mock_order_stream = mock.Mock(
            spec=OrderStream, _stream=mock.Mock(_running=True)
        )
        self.mock_flumine = mock.Mock(
            SIMULATED=False,
            client=mock.Mock(paper_trade=False),
            streams=[self.mock_order_stream],
        )
        self.trading_control = ExecutionValidation(self.mock_flumine)
        self.mock_order = mock.Mock(
            EXCHANGE=ExchangeType.BETFAIR,
            responses=mock.Mock(
                replace_responses=[], update_responses=[], cancel_responses=[]
            ),
        )

    def tearDown(self) -> None:
        config.execution_retry_attempts = 10

    def test_order_stream_connected(self):
        self.assertTrue(self.trading_control.order_stream_connected)
        self.mock_order_stream.stream_running = False
        self.assertFalse(self.trading_control.order_stream_connected)

    def test_order_stream_missing(self):
        self.mock_flumine.streams = []
        self.assertFalse(self.trading_control.order_stream_connected)

    def test_failed_execution_attempts(self):
        mock_responses = [
            mock.Mock(status="SUCCESS"),
            mock.Mock(status="SUCCESS"),
            mock.Mock(status="SUCCESS"),
        ]
        self.assertEqual(
            self.trading_control.failed_execution_attempts(mock_responses), 0
        )

        mock_responses = [
            mock.Mock(status="FAILURE"),
            mock.Mock(status="SUCCESS"),
            mock.Mock(status="TIMEOUT"),
        ]
        self.assertEqual(
            self.trading_control.failed_execution_attempts(mock_responses), 2
        )

    @mock.patch("flumine.controls.tradingcontrols.ExecutionValidation._on_error")
    def test_validate_order(self, mock_on_error):
        self.trading_control.validate_order(self.mock_order, OrderPackageType.PLACE)
        mock_on_error.assert_not_called()

        self.trading_control.validate_order(self.mock_order, OrderPackageType.REPLACE)
        mock_on_error.assert_not_called()

        self.trading_control.validate_order(self.mock_order, OrderPackageType.UPDATE)
        mock_on_error.assert_not_called()

        self.trading_control.validate_order(self.mock_order, OrderPackageType.CANCEL)
        mock_on_error.assert_not_called()

    @mock.patch(
        "flumine.controls.tradingcontrols.ExecutionValidation.failed_execution_attempts"
    )
    def test_validate_order_responses(self, mock_failed_execution_attempts):
        self.mock_replace_responses = [mock.Mock()]
        self.mock_order.responses.replace_responses = self.mock_replace_responses
        self.trading_control.validate_order(self.mock_order, OrderPackageType.REPLACE)
        mock_failed_execution_attempts.assert_called_with(self.mock_replace_responses)

        mock_failed_execution_attempts.reset_mock()

        self.mock_update_responses = [mock.Mock()]
        self.mock_order.responses.update_responses = self.mock_update_responses
        self.trading_control.validate_order(self.mock_order, OrderPackageType.UPDATE)
        mock_failed_execution_attempts.assert_called_with(self.mock_update_responses)

        mock_failed_execution_attempts.reset_mock()

        self.mock_cancel_responses = [mock.Mock()]
        self.mock_order.responses.cancel_responses = self.mock_cancel_responses
        self.trading_control.validate_order(self.mock_order, OrderPackageType.CANCEL)
        mock_failed_execution_attempts.assert_called_with(self.mock_cancel_responses)

    @mock.patch("flumine.controls.tradingcontrols.ExecutionValidation._on_error")
    @mock.patch(
        "flumine.controls.tradingcontrols.ExecutionValidation.order_stream_connected",
        new_callable=mock.PropertyMock,
    )
    @mock.patch(
        "flumine.controls.tradingcontrols.ExecutionValidation.failed_execution_attempts"
    )
    def test_validate_error(
        self, mock_failed_execution_attempts, mock_order_stream_connected, mock_on_error
    ):
        mock_failed_execution_attempts.return_value = 5
        mock_order_stream_connected.return_value = False
        self.trading_control.validate_order(self.mock_order, OrderPackageType.CANCEL)
        mock_on_error.assert_not_called()

        mock_failed_execution_attempts.return_value = 12
        mock_order_stream_connected.return_value = True
        self.trading_control.validate_order(self.mock_order, OrderPackageType.CANCEL)
        mock_on_error.assert_not_called()

        mock_failed_execution_attempts.return_value = 12
        mock_order_stream_connected.return_value = False
        self.trading_control.validate_order(self.mock_order, OrderPackageType.CANCEL)
        mock_on_error.assert_called()

    @mock.patch("flumine.controls.tradingcontrols.ExecutionValidation._on_error")
    @mock.patch(
        "flumine.controls.tradingcontrols.ExecutionValidation.order_stream_connected",
        new_callable=mock.PropertyMock,
    )
    @mock.patch(
        "flumine.controls.tradingcontrols.ExecutionValidation.failed_execution_attempts"
    )
    def test__validate_execution_retry_attempts(
        self, mock_failed_execution_attempts, mock_order_stream_connected, mock_on_error
    ):
        mock_failed_execution_attempts.return_value = 5
        mock_order_stream_connected.return_value = False
        config.execution_retry_attempts = 4
        self.trading_control.validate_order(self.mock_order, OrderPackageType.CANCEL)
        mock_on_error.assert_called()

    @mock.patch("flumine.controls.tradingcontrols.ExecutionValidation.validate_order")
    def test__validate(self, mock_validate_order):
        self.mock_flumine.clients.simulated = False
        self.trading_control._validate(self.mock_order, OrderPackageType.CANCEL)
        mock_validate_order.assert_called()

        mock_validate_order.reset_mock()

        self.mock_flumine.clients.simulated = True
        self.trading_control._validate(self.mock_order, OrderPackageType.CANCEL)
        mock_validate_order.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.ExecutionValidation.validate_order")
    def test__validate_exchange(self, mock_validate_order):
        self.mock_order.EXCHANGE = "OTHER"
        self.trading_control._validate(self.mock_order, OrderPackageType.CANCEL)
        mock_validate_order.assert_not_called()


class TestStrategyExposure(unittest.TestCase):
    def setUp(self):
        self.market = mock.Mock()
        self.market.market_book = mock.Mock(
            number_of_active_runners=6, number_of_winners=1
        )
        self.market.blotter = Blotter("market_id")
        self.mock_flumine = mock.Mock()
        self.mock_flumine.markets.markets = {"market_id": self.market}
        self.trading_control = StrategyExposure(self.mock_flumine)

    def test_init(self):
        self.assertEqual(self.trading_control.NAME, "STRATEGY_EXPOSURE")
        self.assertEqual(self.trading_control.flumine, self.mock_flumine)

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_strategy_validate_order(self, mock_on_error):
        mock_order = mock.Mock(market_id="market_id", lookup=(1, 2, 3))
        mock_order.trade.strategy.validate_order.return_value = True
        mock_runner_context = mock.Mock()
        mock_order.trade.strategy.get_runner_context.return_value = mock_runner_context
        self.trading_control._validate(mock_order, OrderPackageType.PLACE)

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_strategy_validate_order_error(self, mock_on_error):
        mock_order = mock.Mock(market_id="market_id", lookup=(1, 2, 3))
        mock_order.trade.strategy.validate_order.return_value = False
        mock_runner_context = mock.Mock()
        mock_order.trade.strategy.get_runner_context.return_value = mock_runner_context
        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        mock_on_error.assert_called_with(
            mock_order,
            mock_order.violation_msg,
        )

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_limit(self, mock_on_error):
        mock_order = mock.Mock(
            market_id="market_id", lookup=(1, 2, 3), EXCHANGE=ExchangeType.BETFAIR
        )
        mock_order.trade.strategy.max_order_exposure = 10
        mock_order.trade.strategy.max_selection_exposure = 100
        mock_order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        mock_order.order_type.price_ladder_definition = "CLASSIC"
        mock_order.side = "BACK"
        mock_order.order_type.size = 12.0
        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        mock_on_error.assert_called_with(
            mock_order,
            "Order exposure (12.0) is greater than strategy.max_order_exposure (10)",
        )

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_limit_target(self, mock_on_error):
        mock_order = mock.Mock(
            market_id="market_id", lookup=(1, 2, 3), EXCHANGE=ExchangeType.BETFAIR
        )
        mock_order.trade.strategy.max_order_exposure = 10
        mock_order.trade.strategy.max_selection_exposure = 100
        mock_order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        mock_order.order_type.price_ladder_definition = "CLASSIC"
        mock_order.side = "BACK"
        mock_order.order_type.size = None
        mock_order.order_type.bet_target_size = 12.0
        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        mock_on_error.assert_called_with(
            mock_order,
            "Order exposure (12.0) is greater than strategy.max_order_exposure (10)",
        )

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_limit_with_multiple_strategies_succeeds(self, mock_on_error):
        """
        The 2 orders would exceed selection_exposure limits if they were for the same strategy. But they are
        for different strategies. Assert that they do not cause a validation failure.
        """
        strategy = mock.Mock()
        strategy.max_order_exposure = 10
        strategy.max_selection_exposure = 10

        order1 = mock.Mock(market_id="market_id", lookup=(1, 2, 3))
        order1.trade.strategy.max_order_exposure = 10
        order1.trade.strategy.max_selection_exposure = 10
        order1.trade.strategy.max_market_exposure = None
        order1.order_type.ORDER_TYPE = OrderTypes.LIMIT
        order1.order_type.price_ladder_definition = "CLASSIC"
        order1.side = "BACK"
        order1.order_type.price = 2.0
        order1.order_type.size = 9.0
        order1.size_remaining = 9.0
        order1.average_price_matched = 0.0
        order1.size_matched = 0

        order2 = mock.Mock(lookup=(1, 2, 3))
        order2.trade.strategy.max_order_exposure = 10
        order2.trade.strategy.max_selection_exposure = 10
        order2.trade.strategy.max_market_exposure = None
        order2.trade.strategy = strategy
        order2.order_type.ORDER_TYPE = OrderTypes.LIMIT
        order2.order_type.price_ladder_definition = "CLASSIC"
        order2.side = "BACK"
        order2.order_type.price = 3.0
        order2.order_type.size = 9.0
        order2.size_remaining = 5.0
        order2.average_price_matched = 0.0
        order2.size_matched = 0

        self.market.blotter._orders = {"order1": order1, "order2": order2}

        self.trading_control._validate(order1, OrderPackageType.PLACE)
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_limit_with_multiple_strategies_fails(self, mock_on_error):
        """
        The 2 orders are from the same strategy. And are each less than strategy.max_order_exposure.
        However, in combination, they exceed strategy.max_selection_exposure.
        """
        strategy = mock.Mock()
        strategy.max_order_exposure = 10
        strategy.max_selection_exposure = 10

        order1 = mock.Mock(
            market_id="market_id",
            lookup=(1, 2, 3),
            selection_id=2,
            handicap=3,
            side="BACK",
            size_remaining=9.0,
            average_price_matched=0.0,
            size_matched=0,
            status=OrderStatus.EXECUTABLE,
            complete=False,
            EXCHANGE=ExchangeType.BETFAIR,
        )
        order1.trade.strategy = strategy
        order1.order_type.ORDER_TYPE = OrderTypes.LIMIT
        order1.order_type.price_ladder_definition = "CLASSIC"
        order1.order_type.price = 2.0
        order1.order_type.size = 9.0

        order2 = mock.Mock(
            lookup=(1, 2, 3),
            selection_id=2,
            handicap=3,
            side="BACK",
            size_remaining=5.0,
            average_price_matched=0.0,
            size_matched=0,
            status=OrderStatus.EXECUTABLE,
            complete=False,
            EXCHANGE=ExchangeType.BETFAIR,
        )
        order2.trade.strategy = strategy
        order2.order_type.ORDER_TYPE = OrderTypes.LIMIT
        order2.order_type.price_ladder_definition = "CLASSIC"
        order2.order_type.price = 3.0
        order2.order_type.size = 9.0

        self.market.blotter["order2"] = order2

        self.trading_control._validate(order1, OrderPackageType.PLACE)
        self.assertEqual(1, mock_on_error.call_count)
        mock_on_error.assert_called_with(
            order1,
            "Potential selection exposure (14.00) is greater than strategy.max_selection_exposure (10)",
        )

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_limit_line_range(self, mock_on_error):
        mock_order = mock.Mock(
            market_id="market_id", lookup=(1, 2, 3), EXCHANGE=ExchangeType.BETFAIR
        )
        mock_order.trade.strategy.max_order_exposure = 10
        mock_order.trade.strategy.max_selection_exposure = 100
        mock_order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        mock_order.order_type.price_ladder_definition = "LINE_RANGE"
        mock_order.side = "LAY"
        mock_order.order_type.size = 12.0
        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        mock_on_error.assert_called_with(
            mock_order,
            "Order exposure (12.0) is greater than strategy.max_order_exposure (10)",
        )

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_limit_on_close(self, mock_on_error):
        mock_order = mock.Mock(
            market_id="market_id", lookup=(1, 2, 3), EXCHANGE=ExchangeType.BETFAIR
        )
        mock_order.trade.strategy.max_order_exposure = 10
        mock_order.trade.strategy.max_selection_exposure = 100
        mock_order.order_type.ORDER_TYPE = OrderTypes.LIMIT_ON_CLOSE
        mock_order.order_type.price_ladder_definition = "CLASSIC"
        mock_order.side = "BACK"
        mock_order.order_type.liability = 12
        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        mock_on_error.assert_called_with(
            mock_order,
            "Order exposure (12) is greater than strategy.max_order_exposure (10)",
        )

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_market_on_close(self, mock_on_error):
        mock_market = mock.Mock()
        mock_market.blotter.selection_exposure.return_value = 10.0
        self.mock_flumine.markets.markets = {"1.234": mock_market}

        mock_order = mock.Mock(
            market_id="1.234", lookup=(1, 2, 3), EXCHANGE=ExchangeType.BETFAIR
        )
        mock_order.trade.strategy.max_order_exposure = 10
        mock_order.trade.strategy.max_selection_exposure = 100
        mock_order.order_type.ORDER_TYPE = OrderTypes.MARKET_ON_CLOSE
        mock_order.side = "BACK"
        mock_order.order_type.liability = 12
        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        mock_on_error.assert_called_with(
            mock_order,
            "Order exposure (12) is greater than strategy.max_order_exposure (10)",
        )

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_selection(self, mock_on_error):
        mock_market = mock.Mock()
        mock_market.blotter.get_exposures.return_value = {
            "worst_possible_profit_on_win": -12.0
        }
        self.mock_flumine.markets.markets = {"1.234": mock_market}
        mock_order = mock.Mock(
            market_id="1.234",
            lookup=(1, 2, 3),
            side="LAY",
            EXCHANGE=ExchangeType.BETFAIR,
        )
        mock_order.trade.strategy.max_order_exposure = 10
        mock_order.trade.strategy.max_selection_exposure = 10
        mock_order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        mock_order.order_type.price_ladder_definition = "CLASSIC"
        mock_order.order_type.size = 12.0
        mock_order.order_type.price = 1.01
        mock_market.blotter._live_orders = [mock_order]
        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        mock_on_error.assert_called_with(
            mock_order,
            "Potential selection exposure (12.12) is greater than strategy.max_selection_exposure (10)",
        )

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_selection2a(self, mock_on_error):
        """
        test_validate_selection2a expects an error, as it attempts to place a lay bet with a £36 potential
        loss, which is more than the £10 max_selection_exposure
        """
        mock_market = mock.Mock()
        mock_market.blotter = Blotter(market_id="1.234")
        self.mock_flumine.markets.markets = {"1.234": mock_market}

        mock_strategy = mock.Mock()
        mock_strategy.max_order_exposure = 100
        mock_strategy.max_selection_exposure = 10

        mock_trade = mock.Mock()
        mock_trade.strategy = mock_strategy

        mock_order = mock.Mock(
            market_id="1.234",
            lookup=(1, 2, 3),
            side="LAY",
            EXCHANGE=ExchangeType.BETFAIR,
        )
        mock_order.trade = mock_trade
        mock_order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        mock_order.order_type.price_ladder_definition = "CLASSIC"
        mock_order.order_type.size = 9.0
        mock_order.order_type.price = 5.0

        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        mock_on_error.assert_called_with(
            mock_order,
            "Potential selection exposure (36.00) is greater than strategy.max_selection_exposure (10)",
        )

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_selection2b(self, mock_on_error):
        """
        test_validate_selection2b expects no error.
        Unlike test_validate_selection2a, the blotter contains an existing order. The order that it attempts
        to validate hedges the existing order, and reduces the total exposure.
        """
        mock_market = mock.Mock()
        mock_market.market_book = mock.Mock(
            number_of_active_runners=6, number_of_winners=1
        )
        mock_market.blotter = Blotter(market_id="1.234")

        self.mock_flumine.markets.markets = {"1.234": mock_market}

        mock_strategy = mock.Mock()
        mock_strategy.max_order_exposure = 100
        mock_strategy.max_selection_exposure = 10
        mock_strategy.max_market_exposure = None

        mock_trade = mock.Mock()
        mock_trade.strategy = mock_strategy

        existing_matched_order = mock.Mock(
            market_id="1.234", lookup=(1, 2, 3), side="BACK", selection_id=2, handicap=3
        )
        existing_matched_order.trade = mock_trade
        existing_matched_order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        existing_matched_order.size_matched = 9.0
        existing_matched_order.average_price_matched = 6.0
        existing_matched_order.size_remaining = 0.0

        mock_order = mock.Mock(
            market_id="1.234", lookup=(1, 2, 3), side="LAY", selection_id=2, handicap=3
        )
        mock_order.trade = mock_trade
        mock_order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        mock_order.order_type.price_ladder_definition = "CLASSIC"
        mock_order.order_type.size = 9.0
        mock_order.order_type.price = 5.0
        mock_order.size_matched = 0.0

        mock_market.blotter["existing_order"] = existing_matched_order

        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        mock_on_error.assert_not_called()

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_replace(self, mock_on_error):
        """
        Check that validating a REPLACE order does not lead to double counting of exposures.

        In this test, max_selection_exposure is 10.0, and the potential liability on the order is £9.
        If exposures are double counted, the validation would fail.
        If exposures aren't double counted, then the validation will succeed
        """
        strategy = mock.Mock()
        strategy.max_order_exposure = 10
        strategy.max_selection_exposure = 10
        strategy.max_market_exposure = None

        order1 = mock.Mock(
            market_id="market_id",
            lookup=(1, 2, 3),
            selection_id=1234,
            side="BACK",
            average_price_matched=0.0,
            size_matched=0,
            handicap=0,
            status=OrderStatus.EXECUTABLE,
            complete=False,
            EXCHANGE=ExchangeType.BETFAIR,
        )
        order1.trade.strategy = strategy
        order1.order_type.ORDER_TYPE = OrderTypes.LIMIT
        order1.order_type.price_ladder_definition = "CLASSIC"
        order1.order_type.price = 2.0
        order1.order_type.size = 9.0
        order1.size_remaining = 9.0

        self.market.blotter._strategy_selection_orders = {(strategy, 2, 3): [order1]}

        # Show that the exposures aren't double counted when REPLACE is used
        self.trading_control._validate(order1, OrderPackageType.REPLACE)
        mock_on_error.assert_not_called()

        # Just to be sure, check that the validation fails if we try to validate order1 as a PLACE
        self.trading_control._validate(order1, OrderPackageType.PLACE)
        mock_on_error.assert_called_once()

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_limit_betdaq(self, mock_on_error):
        mock_order = mock.Mock(
            market_id="market_id", lookup=(1, 2, 3), EXCHANGE=ExchangeType.BETDAQ
        )
        mock_order.trade.strategy.max_order_exposure = 10
        mock_order.trade.strategy.max_selection_exposure = 100
        mock_order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        mock_order.side = "BACK"
        mock_order.order_type.size = 12.0
        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        mock_on_error.assert_called_with(
            mock_order,
            "Order exposure (12.0) is greater than strategy.max_order_exposure (10)",
        )

    @mock.patch("flumine.controls.tradingcontrols.StrategyExposure._on_error")
    def test_validate_max_market_exposure(self, mock_on_error):
        mock_market = mock.Mock()
        mock_market.market_book = mock.Mock(
            number_of_active_runners=7, number_of_winners=1
        )
        mock_market.blotter = Blotter(market_id="1.240721772")

        self.mock_flumine.markets.markets = {"1.240721772": mock_market}

        mock_strategy = mock.Mock()
        mock_strategy.max_order_exposure = None
        mock_strategy.max_selection_exposure = None
        mock_strategy.max_market_exposure = 30

        for customer_order_ref, (selection_id, price) in enumerate(
            (
                (26456716, 11.5),
                (42699494, 9.4),
                (62698391, 14),
                (42699494, 9.4),
                (26456716, 11),
                (42699494, 9.2),
                (62698391, 13.5),
                (62698391, 13),
                (26456716, 10.5),
                (26456716, 10.5),
            )
        ):
            mock_market.blotter[customer_order_ref] = mock.Mock(
                market_id="1.240721772",
                lookup=("1.240721772", selection_id, 0),
                selection_id=selection_id,
                side="BACK",
                average_price_matched=price,
                size_matched=2,
                handicap=0,
                status=OrderStatus.EXECUTABLE,
                complete=False,
                EXCHANGE=ExchangeType.BETFAIR,
                order_type=mock.Mock(
                    ORDER_TYPE=OrderTypes.LIMIT,
                    price_ladder_definition="CLASSIC",
                    price=price,
                    size=2,
                ),
                size_remaining=0,
                trade=mock.Mock(strategy=mock_strategy),
            )

        mock_order = mock.Mock(
            market_id="1.240721772",
            lookup=("1.240721772", 42699494, 0),
            selection_id=42699494,
            side="BACK",
            average_price_matched=0.0,
            size_matched=0,
            handicap=0,
            status=OrderStatus.EXECUTABLE,
            complete=False,
            EXCHANGE=ExchangeType.BETFAIR,
            order_type=mock.Mock(
                ORDER_TYPE=OrderTypes.LIMIT,
                price_ladder_definition="CLASSIC",
                price=9.6,
                size=2,
            ),
            size_remaining=2,
            trade=mock.Mock(strategy=mock_strategy),
        )

        # 1. Order does not exceed max market exposure
        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        mock_on_error.assert_not_called()

        mock_strategy.max_market_exposure = 20

        # 2. Order replaces an existing order so does not exceed max market exposure
        self.trading_control._validate(mock_order, OrderPackageType.REPLACE)
        mock_on_error.assert_not_called()

        # 3. Order exceeds max market exposure
        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        self.assertEqual(1, mock_on_error.call_count)
        mock_on_error.assert_called_with(
            mock_order,
            "Potential market exposure (22.00) is greater than strategy.max_market_exposure (20)",
        )

    def test_optional_max_exposures(self):
        mock_market = mock.Mock()
        # Test blotter is not accessed
        del mock_market.blotter
        mock_order = mock.Mock(
            lookup=(1, 2, 3),
            market_id="1.234",
            trade=mock.Mock(
                strategy=mock.Mock(
                    max_order_exposure=None,
                    max_selection_exposure=None,
                    max_market_exposure=None,
                )
            ),
        )
        # Test order type is not accessed
        del mock_order.order_type
        self.mock_flumine.markets.markets = {"1.234": mock_market}

        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        self.trading_control._validate(mock_order, OrderPackageType.REPLACE)



import logging
from enum import Enum

from ..exceptions import ClientError

logger = logging.getLogger(__name__)


class ExchangeType(Enum):
    BETFAIR = "Betfair"
    SIMULATED = "Simulated"
    BETCONNECT = "BetConnect"
    BETDAQ = "BETDAQ"


class Clients:
    """
    Data structure for flumine clients.
    First client added == default client
    """

    def __init__(self) -> None:
        self._clients = []
        self._exchange_clients = {exchange: {} for exchange in ExchangeType}

    def add_client(self, client):
        if client in self._clients:
            raise ClientError("Client already present")
        if client.EXCHANGE not in self._exchange_clients:
            raise ClientError("Unknown exchange type")
        exchange = self._exchange_clients[client.EXCHANGE]
        if client.username in exchange:
            raise ClientError(
                "Client username '{0}' already present in '{1}'".format(
                    client.username, client.EXCHANGE
                )
            )
        self._clients.append(client)
        exchange[client.username] = client
        logger.info("Client added", extra=client.info)
        return client

    def get_client(self, exchange_type: ExchangeType, username: str):
        try:
            return self._exchange_clients[exchange_type][username]
        except KeyError:
            return

    def get_default(self):
        return self._clients[0]

    def get_betfair_default(self):
        for client in self._clients:
            if client.EXCHANGE == ExchangeType.BETFAIR:
                return client

    def get_betdaq_default(self):
        for client in self._clients:
            if client.EXCHANGE == ExchangeType.BETDAQ:
                return client

    def login(self) -> None:
        for client in self._clients:
            client.login()
            logger.info("Client login", extra=client.info)

    def keep_alive(self) -> None:
        for client in self._clients:
            client.keep_alive()
            logger.info("Client keep alive", extra=client.info)

    def logout(self) -> None:
        for client in self._clients:
            client.logout()
            logger.info("Client logout", extra=client.info)

    def update_account_details(self) -> None:
        for client in self._clients:
            client.update_account_details()
            logger.info("Client update account details", extra=client.info)

    @property
    def simulated(self) -> bool:
        # return True if simulated client present
        for client in self._clients:
            if client.EXCHANGE == ExchangeType.SIMULATED or client.paper_trade:
                return True
        return False

    @property
    def info(self) -> dict:
        return {
            k.value: {k_i: v_i.info for k_i, v_i in v.items()}
            for k, v in self._exchange_clients.items()
        }

    def __iter__(self):
        return iter(self._clients)

    def __len__(self) -> int:
        return len(self._clients)



import unittest

from flumine import config


class ConfigTest(unittest.TestCase):
    def test_init(self):
        self.assertFalse(config.simulated)
        self.assertTrue(config.simulated_strategy_isolation)
        self.assertFalse(config.simulation_available_prices)
        self.assertIsNone(config.customer_strategy_ref)
        self.assertIsInstance(config.process_id, int)
        self.assertIsNone(config.current_time)
        self.assertFalse(config.raise_errors)
        self.assertEqual(config.max_execution_workers, 32)
        self.assertFalse(config.async_place_orders)
        self.assertEqual(config.place_latency, 0.120)
        self.assertEqual(config.cancel_latency, 0.170)
        self.assertEqual(config.update_latency, 0.150)
        self.assertEqual(config.replace_latency, 0.280)
        self.assertEqual(config.order_sep, "-")
        self.assertEqual(config.execution_retry_attempts, 10)



import betdaq
from enum import Enum
from betdaq.enums import OrderKillType, WithdrawRepriceOption
from betfairlightweight.resources.bettingresources import LineRangeInfo

from ..clients.clients import ExchangeType


class OrderTypes(Enum):
    LIMIT = "Limit"
    LIMIT_ON_CLOSE = "Limit on close"
    MARKET_ON_CLOSE = "Market on close"


class BaseOrderType:
    EXCHANGE = None
    ORDER_TYPE = None

    def place_instruction(self) -> dict:
        raise NotImplementedError

    @property
    def info(self):
        raise NotImplementedError


class LimitOrder(BaseOrderType):
    EXCHANGE = ExchangeType.BETFAIR
    ORDER_TYPE = OrderTypes.LIMIT

    def __init__(
        self,
        price: float,
        size: float = None,
        persistence_type: str = "LAPSE",
        time_in_force: str = None,
        min_fill_size: float = None,
        bet_target_type: str = None,
        bet_target_size: float = None,
        price_ladder_definition: str = "CLASSIC",
        line_range_info: LineRangeInfo = None,
    ):
        self.price = price
        self.size = size
        self.persistence_type = persistence_type
        self.time_in_force = time_in_force
        self.min_fill_size = min_fill_size
        self.bet_target_type = bet_target_type
        self.bet_target_size = bet_target_size
        self.price_ladder_definition = price_ladder_definition
        self.line_range_info = line_range_info

    def place_instruction(self) -> dict:
        return {
            "size": self.size,
            "price": self.price,
            "persistenceType": self.persistence_type,
            "timeInForce": self.time_in_force,
            "minFillSize": self.min_fill_size,
            "betTargetType": self.bet_target_type,
            "betTargetSize": self.bet_target_size,
        }

    @property
    def info(self):
        return {
            "order_type": self.ORDER_TYPE.value,
            "price": self.price,
            "size": self.size,
            "persistence_type": self.persistence_type,
            "time_in_force": self.time_in_force,
            "min_fill_size": self.min_fill_size,
            "bet_target_type": self.bet_target_type,
            "bet_target_size": self.bet_target_size,
            "price_ladder_definition": self.price_ladder_definition,
        }


class LimitOnCloseOrder(BaseOrderType):
    EXCHANGE = ExchangeType.BETFAIR
    ORDER_TYPE = OrderTypes.LIMIT_ON_CLOSE

    def __init__(
        self, liability: float, price: float, price_ladder_definition: str = "CLASSIC"
    ):
        self.liability = liability
        self.price = price
        self.price_ladder_definition = price_ladder_definition

    def place_instruction(self) -> dict:
        return {
            "liability": self.liability,
            "price": self.price,
        }

    @property
    def info(self):
        return {
            "order_type": self.ORDER_TYPE.value,
            "liability": self.liability,
            "price": self.price,
            "price_ladder_definition": self.price_ladder_definition,
        }


class MarketOnCloseOrder(BaseOrderType):
    EXCHANGE = ExchangeType.BETFAIR
    ORDER_TYPE = OrderTypes.MARKET_ON_CLOSE

    def __init__(self, liability: float):
        self.liability = liability

    def place_instruction(self) -> dict:
        return {
            "liability": self.liability,
        }

    @property
    def info(self):
        return {
            "order_type": self.ORDER_TYPE.value,
            "liability": self.liability,
        }


class BetdaqLimitOrder(BaseOrderType):
    EXCHANGE = ExchangeType.BETDAQ
    ORDER_TYPE = OrderTypes.LIMIT

    def __init__(
        self,
        price,
        size,
        betdaq_runner_id,
        runner_reset_count,
        withdrawal_sequence_number,
        kill_type=OrderKillType.FillOrKillDontCancel.value,
        fill_or_kill_threshold=0.0,
        cancel_on_in_running=True,
        cancel_if_selection_reset=True,
        withdrawal_reprice_option=WithdrawRepriceOption.Cancel.value,
    ):
        self.price = price
        self.size = size
        self.betdaq_runner_id = betdaq_runner_id
        self.runner_reset_count = runner_reset_count
        self.withdrawal_sequence_number = withdrawal_sequence_number
        self.kill_type = kill_type
        self.fill_or_kill_threshold = fill_or_kill_threshold
        self.cancel_on_in_running = cancel_on_in_running
        self.cancel_if_selection_reset = cancel_if_selection_reset
        self.withdrawal_reprice_option = withdrawal_reprice_option
        self.price_ladder_definition = None

    def place_instruction(self, polarity, ref=None) -> dict:
        return betdaq.filters.create_order(
            SelectionId=self.betdaq_runner_id,
            Stake=self.size,
            Price=self.price,
            Polarity=polarity,
            ExpectedSelectionResetCount=self.runner_reset_count,
            ExpectedWithdrawalSequenceNumber=self.withdrawal_sequence_number,
            KillType=self.kill_type,
            FillOrKillThreshold=self.fill_or_kill_threshold,
            CancelOnInRunning=self.cancel_on_in_running,
            CancelIfSelectionReset=self.cancel_if_selection_reset,
            WithdrawalRepriceOption=self.withdrawal_reprice_option,
            PunterReferenceNumber=ref,
        )

    @property
    def info(self):
        return {
            "order_type": self.ORDER_TYPE,
            "price": self.price,
            "size": self.size,
            "betdaq_runner_id": self.betdaq_runner_id,
            "runner_reset_count": self.runner_reset_count,
            "withdrawal_sequence_number": self.withdrawal_sequence_number,
            "kill_type": self.kill_type,
            "fill_or_kill_threshold": self.fill_or_kill_threshold,
            "cancel_on_in_running": self.cancel_on_in_running,
            "cancel_if_selection_reset": self.cancel_if_selection_reset,
            "withdrawal_reprice_option": self.withdrawal_reprice_option,
        }



import unittest
from unittest import mock

from flumine.events import events


class BaseEventTest(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_event = mock.Mock()
        self.base_event = events.BaseEvent(self.mock_event, 123)

    def test_init(self):
        self.assertIsNone(self.base_event.EVENT_TYPE)
        self.assertIsNone(self.base_event.QUEUE_TYPE)
        self.assertEqual(self.base_event.event, self.mock_event)
        self.assertEqual(self.base_event.exchange, 123)
        self.assertIsNotNone(self.base_event._time_created)

    def test_elapsed_seconds(self):
        self.assertGreaterEqual(self.base_event.elapsed_seconds, 0)

    def test_str(self):
        self.base_event = events.MarketBookEvent(None)
        self.assertEqual(str(self.base_event), "<MARKET_BOOK [HANDLER]>")



import logging
import datetime
from typing import Optional

logger = logging.getLogger(__name__)


class RunnerContext:
    """Runner context at strategy level"""

    def __init__(self, selection_id: int):
        self.selection_id = selection_id
        self.invested = False
        self.datetime_last_placed = None
        self.datetime_last_reset = None
        self.trades = []
        self.live_trades = []

    def place(self, trade_id) -> None:
        self.invested = True
        self.datetime_last_placed = datetime.datetime.now(datetime.timezone.utc)
        if trade_id not in self.trades:
            self.trades.append(trade_id)
        if trade_id not in self.live_trades:
            self.live_trades.append(trade_id)

    def reset(self, trade_id) -> None:
        self.datetime_last_reset = datetime.datetime.now(datetime.timezone.utc)
        try:
            self.live_trades.remove(trade_id)
        except ValueError:
            logger.warning(
                "Trade '%s' not present in RunnerContext live_trades on reset"
                % trade_id
            )

    @property
    def executable_orders(self) -> bool:
        if self.live_trades:
            return True
        else:
            return False

    @property
    def trade_count(self) -> int:
        return len(self.trades)

    @property
    def live_trade_count(self) -> int:
        return len(self.live_trades)

    @property
    def placed_elapsed_seconds(self) -> Optional[float]:
        if self.datetime_last_placed:
            return (
                datetime.datetime.now(datetime.timezone.utc) - self.datetime_last_placed
            ).total_seconds()

    @property
    def reset_elapsed_seconds(self) -> Optional[float]:
        if self.datetime_last_reset:
            return (
                datetime.datetime.now(datetime.timezone.utc) - self.datetime_last_reset
            ).total_seconds()



import unittest
from unittest import mock
from betfairlightweight.metadata import currency_parameters
from betfairlightweight.exceptions import BetfairError
from betconnect.exceptions import BetConnectException

from flumine.clients.clients import ExchangeType, Clients
from flumine.clients import (
    BaseClient,
    BetfairClient,
    SimulatedClient,
    BetConnectClient,
    BetdaqClient,
)
from flumine.clients import betfairclient
from flumine import exceptions


class ClientsTest(unittest.TestCase):
    def setUp(self):
        self.clients = Clients()

    def test_exchange_type(self):
        self.assertEqual(len(ExchangeType), 4)
        assert ExchangeType

    def test_init(self):
        self.assertEqual(self.clients._clients, [])
        self.assertEqual(
            self.clients._exchange_clients,
            {exchange: {} for exchange in ExchangeType},
        )

    def test_add_client(self):
        mock_client = mock.Mock(EXCHANGE=ExchangeType.BETFAIR, username="test")
        self.clients._clients = [mock_client]
        with self.assertRaises(exceptions.ClientError):
            self.clients.add_client(mock_client)

        with self.assertRaises(exceptions.ClientError):
            self.clients.add_client(mock.Mock(EXCHANGE="test", username="test"))

        self.clients._clients = []
        self.clients._exchange_clients[ExchangeType.BETFAIR][
            mock_client.username
        ] = mock_client
        with self.assertRaises(exceptions.ClientError):
            self.clients.add_client(mock_client)

        self.clients._clients = []
        self.clients._exchange_clients = {exchange: {} for exchange in ExchangeType}
        mock_client.EXCHANGE = ExchangeType.BETFAIR
        self.assertEqual(self.clients.add_client(mock_client), mock_client)
        self.assertEqual(self.clients._clients, [mock_client])
        self.assertEqual(
            self.clients._exchange_clients[mock_client.EXCHANGE],
            {mock_client.username: mock_client},
        )

    def test_get_default(self):
        self.clients._clients.append("howlandthehum")
        self.assertEqual(self.clients.get_default(), "howlandthehum")

    def test_get_betfair_default(self):
        mock_client_one = mock.Mock(EXCHANGE=ExchangeType.SIMULATED)
        mock_client_two = mock.Mock(EXCHANGE=ExchangeType.BETFAIR)
        self.clients._clients.append(mock_client_one)
        self.clients._clients.append(mock_client_two)
        self.assertEqual(self.clients.get_betfair_default(), mock_client_two)

    def test_get_betdaq_default(self):
        mock_client_one = mock.Mock(EXCHANGE=ExchangeType.SIMULATED)
        mock_client_two = mock.Mock(EXCHANGE=ExchangeType.BETDAQ)
        self.clients._clients.append(mock_client_one)
        self.clients._clients.append(mock_client_two)
        self.assertEqual(self.clients.get_betdaq_default(), mock_client_two)

    def test_get_client(self):
        self.clients._exchange_clients[ExchangeType.SIMULATED]["joejames"] = 12
        self.assertEqual(
            self.clients.get_client(ExchangeType.SIMULATED, "joejames"), 12
        )

    def test_login(self):
        mock_client = unittest.mock.Mock()
        self.clients._clients = [mock_client]
        self.clients.login()
        mock_client.login.assert_called_with()

    def test_keep_alive(self):
        mock_client = unittest.mock.Mock()
        self.clients._clients = [mock_client]
        self.clients.keep_alive()
        mock_client.keep_alive.assert_called_with()

    def test_logout(self):
        mock_client = unittest.mock.Mock()
        self.clients._clients = [mock_client]
        self.clients.logout()
        mock_client.logout.assert_called_with()

    def test_update_account_details(self):
        mock_client = unittest.mock.Mock()
        self.clients._clients = [mock_client]
        self.clients.update_account_details()
        mock_client.update_account_details.assert_called_with()

    def test_simulated(self):
        self.assertFalse(self.clients.simulated)
        self.clients._clients.append(mock.Mock(paper_trade=True))
        self.assertTrue(self.clients.simulated)

    def test_info(self):
        self.assertEqual(
            self.clients.info, {exchange.value: {} for exchange in ExchangeType}
        )
        mock_client = mock.Mock()
        self.clients._exchange_clients[ExchangeType.BETFAIR]["james"] = mock_client
        self.assertEqual(
            self.clients.info,
            {
                ExchangeType.BETFAIR.value: {"james": mock_client.info},
                ExchangeType.SIMULATED.value: {},
                ExchangeType.BETCONNECT.value: {},
                ExchangeType.BETDAQ.value: {},
            },
        )

    def test_iter(self):
        for i in self.clients:
            assert i

    def test_len(self):
        self.assertEqual(len(self.clients), 0)


class BaseClientTest(unittest.TestCase):
    def setUp(self):
        self.mock_betting_client = mock.Mock(lightweight=False)
        self.base_client = BaseClient(
            self.mock_betting_client, 1024, 100, 0.02, True, username="test"
        )

    def test_init(self):
        self.assertEqual(self.base_client.betting_client, self.mock_betting_client)
        self.assertEqual(self.base_client.transaction_limit, 1024)
        self.assertEqual(self.base_client.capital_base, 100)
        self.assertEqual(self.base_client.commission_base, 0.02)
        self.assertTrue(self.base_client.interactive_login)
        self.assertEqual(self.base_client._username, "test")
        self.assertIsNone(self.base_client.account_details)
        self.assertIsNone(self.base_client.account_funds)
        self.assertEqual(self.base_client.commission_paid, 0)
        self.assertIsNone(self.base_client.execution)
        self.assertEqual(self.base_client.trading_controls, [])
        self.assertTrue(self.base_client.order_stream)
        self.assertIsNone(self.base_client.order_stream_conflate_ms)
        self.assertTrue(self.base_client.best_price_execution)
        self.assertTrue(self.base_client.min_bet_validation)
        self.assertFalse(self.base_client.paper_trade)
        self.assertFalse(self.base_client.simulated_full_match)
        self.assertIsNone(self.base_client.execution)

    def test_init_assert(self):
        with self.assertRaises(AssertionError):
            BaseClient(mock.Mock(lightweight=True), 1024, 100, 0.02, True)

    def test_login(self):
        with self.assertRaises(NotImplementedError):
            assert self.base_client.login()

    def test_keep_alive(self):
        with self.assertRaises(NotImplementedError):
            assert self.base_client.keep_alive()

    def test_logout(self):
        with self.assertRaises(NotImplementedError):
            assert self.base_client.logout()

    def test_update_account_details(self):
        with self.assertRaises(NotImplementedError):
            assert self.base_client.update_account_details()

    def test_add_execution(self):
        mock_flumine = mock.Mock()
        self.base_client.EXCHANGE = ExchangeType.SIMULATED
        self.base_client.add_execution(mock_flumine)
        self.assertEqual(self.base_client.execution, mock_flumine.simulated_execution)
        self.base_client.EXCHANGE = ExchangeType.BETFAIR
        self.base_client.add_execution(mock_flumine)
        self.assertEqual(self.base_client.execution, mock_flumine.betfair_execution)
        self.base_client.EXCHANGE = ExchangeType.BETDAQ
        self.base_client.add_execution(mock_flumine)
        self.assertEqual(self.base_client.execution, mock_flumine.betdaq_execution)

    def test_add_execution_cls(self):
        mock_flumine = mock.Mock()
        self.base_client.EXCHANGE = ExchangeType.SIMULATED
        mock_execution_cls = mock.Mock()
        self.base_client._execution_cls = mock_execution_cls
        self.base_client.add_execution(mock_flumine)
        self.assertEqual(self.base_client.execution, mock_execution_cls.return_value)
        mock_execution_cls.assert_called_with(mock_flumine)

    def test_add_execution_paper(self):
        self.base_client.paper_trade = True
        self.base_client.EXCHANGE = ExchangeType.BETFAIR
        mock_flumine = mock.Mock()
        self.base_client.add_execution(mock_flumine)
        self.assertEqual(self.base_client.execution, mock_flumine.simulated_execution)

    def test_add_transaction(self):
        mock_trading_control = mock.Mock()
        self.base_client.trading_controls.append(mock_trading_control)
        self.base_client.add_transaction(123, True)
        mock_trading_control.add_transaction.assert_called_with(123, True)

    def test_current_transaction_count_total(self):
        self.assertIsNone(self.base_client.current_transaction_count_total)
        mock_trading_control = mock.Mock(
            NAME="MAX_TRANSACTION_COUNT", current_transaction_count_total=123
        )
        self.base_client.trading_controls.append(mock_trading_control)
        self.assertEqual(self.base_client.current_transaction_count_total, 123)

    def test_transaction_count_total(self):
        self.assertIsNone(self.base_client.transaction_count_total)
        mock_trading_control = mock.Mock(
            NAME="MAX_TRANSACTION_COUNT", transaction_count_total=123
        )
        self.base_client.trading_controls.append(mock_trading_control)
        self.assertEqual(self.base_client.transaction_count_total, 123)

    def test_min_bet_size(self):
        with self.assertRaises(NotImplementedError):
            assert self.base_client.min_bet_size

    def test_min_bet_payout(self):
        with self.assertRaises(NotImplementedError):
            assert self.base_client.min_bet_payout

    def test_min_bsp_liability(self):
        with self.assertRaises(NotImplementedError):
            assert self.base_client.min_bsp_liability

    def test_username(self):
        self.assertEqual(
            self.base_client.username, self.base_client.betting_client.username
        )
        self.base_client.betting_client = None
        self.assertEqual(self.base_client.username, self.base_client._username)

    def test_info(self):
        self.assertTrue(self.base_client.info)


class BetfairClientTest(unittest.TestCase):
    def setUp(self):
        self.mock_betting_client = mock.Mock(lightweight=False)
        self.betfair_client = BetfairClient(self.mock_betting_client)

    def test_init(self):
        self.assertEqual(
            betfairclient.MIN_BET_SIZE, currency_parameters["GBP"]["min_bet_size"]
        )
        self.assertEqual(
            betfairclient.MIN_BSP_LIABILITY,
            currency_parameters["GBP"]["min_bsp_liability"],
        )
        self.assertEqual(
            betfairclient.MIN_BET_PAYOUT, currency_parameters["GBP"]["min_bet_payout"]
        )

    def test_login(self):
        self.betfair_client.login()
        self.mock_betting_client.login.assert_called_with()

    def test_login_no_certs(self):
        self.betfair_client.interactive_login = True
        self.betfair_client.login()
        self.mock_betting_client.login_interactive.assert_called_with()

    def test_login_error(self):
        self.betfair_client.betting_client.login.side_effect = BetfairError
        self.assertIsNone(self.betfair_client.login())
        self.mock_betting_client.login.assert_called_with()

    def test_keep_alive(self):
        self.mock_betting_client.session_expired = True
        self.betfair_client.keep_alive()
        self.mock_betting_client.keep_alive.assert_called_with()

    def test_keep_alive_error(self):
        self.betfair_client.betting_client.keep_alive.side_effect = BetfairError
        self.assertIsNone(self.betfair_client.keep_alive())
        self.mock_betting_client.keep_alive.assert_called_with()

    def test_logout(self):
        self.betfair_client.logout()
        self.mock_betting_client.logout.assert_called_with()

    def test_logout_error(self):
        self.betfair_client.betting_client.logout.side_effect = BetfairError
        self.assertIsNone(self.betfair_client.logout())
        self.mock_betting_client.logout.assert_called_with()

    @mock.patch("flumine.clients.betfairclient.BetfairClient._get_account_details")
    @mock.patch("flumine.clients.betfairclient.BetfairClient._get_account_funds")
    def test_update_account_details(self, mock_get_funds, mock_get_details):
        self.betfair_client.update_account_details()
        mock_get_funds.assert_called_with()
        mock_get_details.assert_called_with()
        self.assertEqual(self.betfair_client.account_details, mock_get_details())
        self.assertEqual(self.betfair_client.account_funds, mock_get_funds())

    def test__get_account_details(self):
        self.betfair_client._get_account_details()
        self.mock_betting_client.account.get_account_details.assert_called_with()

    def test__get_account_details_error(self):
        self.betfair_client.betting_client.account.get_account_details.side_effect = (
            BetfairError
        )
        self.assertIsNone(self.betfair_client._get_account_details())
        self.mock_betting_client.account.get_account_details.assert_called_with()

    def test__get_account_funds(self):
        self.betfair_client._get_account_funds()
        self.mock_betting_client.account.get_account_funds.assert_called_with()

    def test__get_account_funds_error(self):
        self.betfair_client.betting_client.account.get_account_funds.side_effect = (
            BetfairError
        )
        self.assertIsNone(self.betfair_client._get_account_funds())
        self.mock_betting_client.account.get_account_funds.assert_called_with()

    def test_min_bet_size(self):
        mock_account_details = mock.Mock()
        mock_account_details.currency_code = "GBP"
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bet_size, 1)

    def test_min_bet_size_none(self):
        mock_account_details = mock.Mock()
        mock_account_details.currency_code = None
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bet_size, 1)

    def test_min_bet_size_ac_none(self):
        mock_account_details = None
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bet_size, 1)

    def test_min_bsp_liability(self):
        mock_account_details = mock.Mock()
        mock_account_details.currency_code = "USD"
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bsp_liability, 20)

    def test_min_bsp_liability_none(self):
        mock_account_details = mock.Mock()
        mock_account_details.currency_code = None
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bsp_liability, 10)

    def test_min_bsp_liability_ac_none(self):
        mock_account_details = None
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bsp_liability, 10)

    def test_min_bet_payout(self):
        mock_account_details = mock.Mock()
        mock_account_details.currency_code = "GBP"
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bet_payout, 10)

    def test_min_bet_payout_none(self):
        mock_account_details = mock.Mock()
        mock_account_details.currency_code = None
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bet_payout, 10)

    def test_min_bet_payout_ac_none(self):
        mock_account_details = None
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bet_payout, 10)


class SimulatedClientTest(unittest.TestCase):
    def setUp(self):
        self.simulated_client = SimulatedClient()

    def test_login(self):
        self.simulated_client.login()

    def test_keep_alive(self):
        self.simulated_client.keep_alive()

    def test_logout(self):
        self.simulated_client.logout()

    @mock.patch("flumine.clients.simulatedclient.AccountDetails")
    def test_update_account_details(self, mock_account_details):
        self.simulated_client.update_account_details()
        self.assertEqual(self.simulated_client.account_details, mock_account_details())

    def test_min_bet_size(self):
        self.simulated_client.update_account_details()
        self.assertEqual(self.simulated_client.min_bet_size, 1)

    def test_min_bsp_liability(self):
        self.simulated_client.update_account_details()
        self.assertEqual(self.simulated_client.min_bsp_liability, 10)

    def test_min_bet_payout(self):
        self.simulated_client.update_account_details()
        self.assertEqual(self.simulated_client.min_bet_payout, 10)

    def test_username(self):
        self.assertEqual(
            self.simulated_client.username,
            self.simulated_client._username,
        )


class BetConnectClientTest(unittest.TestCase):
    def setUp(self):
        self.mock_betting_client = mock.Mock()
        del self.mock_betting_client.lightweight
        self.betconnect_client = BetConnectClient(self.mock_betting_client)

    def test_login(self):
        self.betconnect_client.login()
        self.mock_betting_client.account.login.assert_called_with()

    def test_login_error(self):
        self.betconnect_client.betting_client.account.login.side_effect = (
            BetConnectException
        )
        self.assertIsNone(self.betconnect_client.login())
        self.mock_betting_client.account.login.assert_called_with()

    def test_keep_alive(self):
        self.mock_betting_client.session_expired = True
        self.betconnect_client.keep_alive()
        self.mock_betting_client.account.refresh_session_token.assert_called_with()

    def test_keep_alive_error(self):
        self.betconnect_client.betting_client.account.refresh_session_token.side_effect = (
            BetConnectException
        )
        self.assertIsNone(self.betconnect_client.keep_alive())
        self.mock_betting_client.account.refresh_session_token.assert_called_with()

    def test_logout(self):
        self.betconnect_client.logout()
        self.mock_betting_client.account.logout.assert_called_with()

    def test_logout_error(self):
        self.betconnect_client.betting_client.account.logout.side_effect = (
            BetConnectException
        )
        self.assertIsNone(self.betconnect_client.logout())
        self.mock_betting_client.account.logout.assert_called_with()

    @mock.patch(
        "flumine.clients.betconnectclient.BetConnectClient._get_account_details"
    )
    @mock.patch("flumine.clients.betconnectclient.BetConnectClient._get_account_funds")
    def test_update_account_details(self, mock_get_funds, mock_get_details):
        self.betconnect_client.update_account_details()
        mock_get_funds.assert_called_with()
        mock_get_details.assert_called_with()
        self.assertEqual(self.betconnect_client.account_details, mock_get_details())
        self.assertEqual(self.betconnect_client.account_funds, mock_get_funds())

    def test__get_account_details(self):
        self.betconnect_client._get_account_details()
        self.mock_betting_client.account.get_user_preferences.assert_called_with()

    def test__get_account_details_error(self):
        self.betconnect_client.betting_client.account.get_user_preferences.side_effect = (
            BetConnectException
        )
        self.assertIsNone(self.betconnect_client._get_account_details())
        self.mock_betting_client.account.get_user_preferences.assert_called_with()

    def test__get_account_funds(self):
        self.betconnect_client._get_account_funds()
        self.mock_betting_client.account.get_balance.assert_called_with()

    def test__get_account_funds_error(self):
        self.betconnect_client.betting_client.account.get_balance.side_effect = (
            BetConnectException
        )
        self.assertIsNone(self.betconnect_client._get_account_funds())
        self.mock_betting_client.account.get_balance.assert_called_with()


class BetdaqClientTest(unittest.TestCase):
    def setUp(self):
        self.mock_betting_client = mock.Mock()
        del self.mock_betting_client.lightweight
        self.betdaq_client = BetdaqClient(self.mock_betting_client)

    def test_login(self):
        self.betdaq_client.login()

    def test_keep_alive(self):
        self.betdaq_client.keep_alive()

    def test_logout(self):
        self.betdaq_client.logout()

    def test_update_account_details(self):
        self.betdaq_client.update_account_details()
        self.assertEqual(
            self.betdaq_client.account_funds,
            self.mock_betting_client.account.get_account_balances.return_value,
        )

    def test_min_bet_size(self):
        self.assertEqual(self.betdaq_client.min_bet_size(), 1)

    def test_min_bet_payout(self):
        self.assertIsNone(self.betdaq_client.min_bet_payout())

    def test_min_bsp_liability(self):
        self.assertIsNone(self.betdaq_client.min_bsp_liability())



import logging
import requests
from typing import Callable
from betdaq import BetdaqError

from ..clients import ExchangeType
from .baseexecution import BaseExecution
from ..exceptions import OrderExecutionError
from ..order.order import BaseOrder
from ..order.orderpackage import BaseOrderPackage, OrderPackageType
from ..events.events import OrderEvent

logger = logging.getLogger(__name__)


class BetdaqExecution(BaseExecution):
    """
    BETDAQ execution for place, cancel and update, default
    single thread pool:

    'A Punter can not issue more than one order API (PlaceSingleOrder,
    PlaceGroupOrder or ChangeOrder) at the same time'
    """

    EXCHANGE = ExchangeType.BETDAQ

    def execute_place(
        self, order_package: BaseOrderPackage, http_session: requests.Session
    ) -> None:
        response = self._execution_helper(self.place, order_package)
        if response:
            order_lookup = {int(o.id): o for o in order_package}
            for instruction_report in response:
                # get order (can't rely on the order they are returned)
                order = order_lookup.pop(instruction_report["customer_reference"])
                with order.trade:
                    self._order_logger(
                        order, instruction_report, order_package.package_type
                    )
                    if not instruction_report["return_code"]:
                        order.executable()  # let process.py pick it up
                    else:
                        logger.error(
                            "execute_place: Order error return code",
                            extra={
                                "order_id": int(order.id),
                                "return_code": instruction_report["return_code"],
                                "response": instruction_report,
                            },
                        )
                        order.execution_complete()
            # todo missing order responses?
            # todo update transaction counts
        else:
            # reset on error so that they can be picked back up
            for order in order_package:
                with order.trade:
                    order.execution_complete()

    def place(self, order_package: BaseOrderPackage):
        return order_package.client.betting_client.betting.place_orders(
            order_list=order_package.place_instructions
        )

    def execute_cancel(
        self, order_package: BaseOrderPackage, http_session: requests.Session
    ) -> None:
        response = self._execution_helper(self.cancel, order_package)
        if response:
            order_lookup = {o.bet_id: o for o in order_package}
            for instruction_report in response:
                # get order (can't rely on the order they are returned)
                order = order_lookup.pop(instruction_report["order_id"])
                with order.trade:
                    self._order_logger(
                        order, instruction_report, order_package.package_type
                    )
                    order.execution_complete()
            # reset any not returned so that they can be picked back up
            for order in order_lookup.values():
                with order.trade:
                    order.executable()
        else:
            # reset on error so that they can be picked back up
            for order in order_package:
                with order.trade:
                    order.executable()

    def cancel(self, order_package: BaseOrderPackage):
        return order_package.client.betting_client.betting.cancel_orders(
            order_ids=order_package.cancel_instructions
        )

    def execute_update(
        self, order_package: BaseOrderPackage, http_session: requests.Session
    ) -> None:
        response = self._execution_helper(self.update, order_package)
        if response:
            order_lookup = {o.bet_id: o for o in order_package}
            for instruction_report in response:
                # get order (can't rely on the order they are returned)
                order = order_lookup.pop(instruction_report["order_id"])
                with order.trade:
                    self._order_logger(
                        order, instruction_report, order_package.package_type
                    )
                    if instruction_report.get("return_code"):
                        logger.error(
                            "execute_update: Order error return code",
                            extra={
                                "bet_id": order.bet_id,
                                "return_code": instruction_report["return_code"],
                                "response": instruction_report,
                            },
                        )
                        order.executable()
                    else:
                        # we don't update the order.status as we don't have a response yet
                        pass
        else:
            # reset on error so that they can be picked back up
            for order in order_package:
                with order.trade:
                    order.executable()

    def update(self, order_package: BaseOrderPackage):
        # temp copy to prevent an empty list of instructions sent
        # this can occur if order is matched during the execution
        order_list = list(order_package.update_instructions)
        if not order_list:
            logger.warning("Empty `order_list`", extra=order_package.info)
            raise OrderExecutionError()
        return order_package.client.betting_client.betting.update_orders(
            order_list=order_list
        )

    def _execution_helper(
        self,
        trading_function: Callable,
        order_package: BaseOrderPackage,
    ):
        try:
            response = trading_function(order_package)
        except BetdaqError as e:
            logger.error(
                "Execution error",
                extra={
                    "trading_function": trading_function.__name__,
                    "response": e,
                    "order_package": order_package.info,
                },
                exc_info=True,
            )
            return
        except Exception as e:
            logger.critical(
                "Execution unknown error",
                extra={
                    "trading_function": trading_function.__name__,
                    "response": e,
                    "order_package": order_package.info,
                },
                exc_info=True,
            )
            return
        logger.info(
            f"execute_{trading_function.__name__}",
            extra={
                "trading_function": trading_function.__name__,
                "response": response,
                "order_package": order_package.info,
            },
        )
        return response

    def _order_logger(
        self, order: BaseOrder, instruction_report: dict, package_type: OrderPackageType
    ):
        logger.info(
            "Order %s: %s",
            package_type.value,
            instruction_report.get("return_code"),
            extra={
                "bet_id": order.bet_id,
                "order_id": order.id,
                "return_code": instruction_report.get("return_code"),
            },
        )
        if package_type == OrderPackageType.PLACE:
            order_id = instruction_report.get("order_id")
            dt = True if order_id else False
            order.responses.placed(instruction_report, dt=dt)
            if order_id:
                order.bet_id = order_id
                self.flumine.log_control(OrderEvent(order, exchange=order.EXCHANGE))
        elif package_type == OrderPackageType.CANCEL:
            order.responses.cancelled(instruction_report)
        elif package_type == OrderPackageType.UPDATE:
            order.responses.updated(instruction_report)



from .clients import ExchangeType, Clients
from .baseclient import BaseClient
from .betfairclient import BetfairClient
from .simulatedclient import SimulatedClient
from .betconnectclient import BetConnectClient
from .betdaqclient import BetdaqClient



import logging
from typing import Optional
from .baseclient import BaseClient
from .clients import ExchangeType

from betdaq.exceptions import BetdaqError

logger = logging.getLogger(__name__)


class BetdaqClient(BaseClient):
    """
    BETDAQ betting client.
    """

    EXCHANGE = ExchangeType.BETDAQ

    def login(self) -> None:
        pass  # NR

    def keep_alive(self) -> None:
        pass  # NR

    def logout(self) -> None:
        pass  # NR

    def update_account_details(self) -> None:
        try:
            self.account_funds = self.betting_client.account.get_account_balances()
            logger.info(
                "account_funds updated",
                extra={
                    "client": repr(self.betting_client),
                    "account_funds": self.account_funds,
                },
            )
        except (BetdaqError, Exception) as e:
            logger.error(
                "BetdaqClient `account.get_account_details` error",
                exc_info=True,
                extra={
                    "client": self.betting_client,
                    "trading_function": "account.get_account_balances",
                    "response": e,
                },
            )

    def min_bet_size(self) -> Optional[float]:
        return 1

    def min_bet_payout(self) -> Optional[float]:
        return None  # todo?

    def min_bsp_liability(self) -> Optional[float]:
        return None  # XSP todo?



import unittest
import datetime
from unittest import mock

from flumine.controls.clientcontrols import (
    BaseControl,
    MaxTransactionCount,
    OrderPackageType,
)
from flumine.exceptions import ControlError


class TestBaseControl(unittest.TestCase):
    def setUp(self):
        self.mock_flumine = mock.Mock()
        self.mock_client = mock.Mock()
        self.control = BaseControl(self.mock_flumine, self.mock_client)

    def test_init(self):
        self.assertEqual(self.control.flumine, self.mock_flumine)
        self.assertIsNone(self.control.NAME)

    @mock.patch("flumine.controls.BaseControl._validate")
    def test_call(self, mock_validate):
        order = mock.Mock()
        self.control(order, OrderPackageType.PLACE)
        mock_validate.assert_called_with(order, OrderPackageType.PLACE)

    def test_validate(self):
        with self.assertRaises(NotImplementedError):
            self.control._validate(None, None)

    def test_on_error(self):
        order = mock.Mock()
        order.info = {"hello": "world"}
        with self.assertRaises(ControlError):
            self.control._on_error(order, "test")
        order.violation.assert_called_with("Order has violated: None Error: test")


class TestMaxTransactionCount(unittest.TestCase):
    def setUp(self):
        self.mock_client = mock.Mock()
        self.mock_client.transaction_limit = 1000
        self.mock_client.chargeable_transaction_count = 0
        self.mock_flumine = mock.Mock()
        self.trading_control = MaxTransactionCount(self.mock_flumine, self.mock_client)

    def test_init(self):
        self.assertEqual(self.trading_control.client, self.mock_client)
        self.assertEqual(self.trading_control.NAME, "MAX_TRANSACTION_COUNT")
        self.assertIsNone(self.trading_control._next_hour)
        self.assertEqual(self.trading_control.current_transaction_count, 0)
        self.assertEqual(self.trading_control.current_failed_transaction_count, 0)
        self.assertEqual(self.trading_control.transaction_count, 0)
        self.assertEqual(self.trading_control.failed_transaction_count, 0)

    def test_add_transaction(self):
        self.trading_control.add_transaction(123)
        self.assertEqual(self.trading_control.transaction_count, 123)
        self.assertEqual(self.trading_control.current_transaction_count, 123)
        self.trading_control.add_transaction(123, True)
        self.assertEqual(self.trading_control.failed_transaction_count, 123)
        self.assertEqual(self.trading_control.current_failed_transaction_count, 123)
        self.trading_control.add_transaction(123)
        self.assertEqual(self.trading_control.transaction_count, 246)
        self.assertEqual(self.trading_control.current_transaction_count, 246)

    @mock.patch("flumine.controls.clientcontrols.MaxTransactionCount._check_hour")
    def test_validate(self, mock_check_hour):
        mock_order = mock.Mock()
        self.trading_control._validate(mock_order, OrderPackageType.PLACE)
        mock_check_hour.assert_called()

    @mock.patch("flumine.controls.clientcontrols.MaxTransactionCount._set_next_hour")
    def test_check_hour(self, mock_set_next_hour):
        self.trading_control._next_hour = datetime.datetime.now(datetime.timezone.utc)
        self.trading_control._check_hour()

        now = datetime.datetime.now()
        self.trading_control._next_hour = (now + datetime.timedelta(hours=-1)).replace(
            minute=0, second=0, microsecond=0
        )
        self.trading_control._check_hour()
        mock_set_next_hour.assert_called_with()

    @mock.patch("flumine.controls.clientcontrols.MaxTransactionCount._set_next_hour")
    def test_check_hour_none(self, mock_set_next_hour):
        self.trading_control._next_hour = None
        self.trading_control._check_hour()
        mock_set_next_hour.assert_called_with()

    def test_check_hour_prev_time(self):
        self.trading_control._set_next_hour()
        self.trading_control.current_transaction_count = 5069
        self.trading_control.current_failed_transaction_count = 5069
        now = datetime.datetime.now(datetime.timezone.utc)
        now_1 = (now + datetime.timedelta(days=1)).replace(
            minute=0, second=0, microsecond=0
        )
        self.trading_control._next_hour = now_1
        self.trading_control._check_hour()

        self.assertEqual(self.trading_control.current_transaction_count, 0)
        self.assertEqual(self.trading_control.current_failed_transaction_count, 0)

    def test_set_next_hour(self):
        self.trading_control.current_transaction_count = 5069
        self.trading_control.current_failed_transaction_count = 5069
        self.trading_control._next_hour = None

        self.trading_control._set_next_hour()
        now = datetime.datetime.now(datetime.timezone.utc)
        now_1 = (now + datetime.timedelta(hours=1)).replace(
            minute=0, second=0, microsecond=0
        )
        self.assertEqual(self.trading_control._next_hour, now_1)
        self.assertEqual(self.trading_control.current_transaction_count, 0)
        self.assertEqual(self.trading_control.current_failed_transaction_count, 0)

    def test_safe(self):
        self.assertTrue(self.trading_control.safe)

        self.trading_control.client.transaction_limit = -10
        self.assertFalse(self.trading_control.safe)

        self.trading_control.client.transaction_limit = None
        self.assertTrue(self.trading_control.safe)

    def test_current_transaction_count_total(self):
        self.trading_control.current_transaction_count = 1
        self.trading_control.current_failed_transaction_count = 2
        self.assertEqual(self.trading_control.current_transaction_count_total, 3)

    def test_transaction_count_total(self):
        self.trading_control.transaction_count = 1
        self.trading_control.failed_transaction_count = 2
        self.assertEqual(self.trading_control.transaction_count_total, 3)

    def test_transaction_limit(self):
        self.assertEqual(
            self.trading_control.transaction_limit, self.mock_client.transaction_limit
        )
